from __future__ import annotations

import base64
import dataclasses
import json
import re
import sys
import typing
import uuid

from ldaptool._utils.dninfo import DNInfo

from . import _types
from .arguments import Arguments

TEntry = dict[str, list[bytes]]
TDecoded = dict[str, list["Attribute"]]

CTRL = re.compile(r"[\x00-\x19]")


@dataclasses.dataclass(slots=True, kw_only=True)
class Attribute:
    name: str
    raw: bytes
    utf8_clean: typing.Optional[str]
    decoded: typing.Optional[str]

    @typing.overload
    def __init__(
        self,
        *,
        name: str,
        raw: bytes,
        arguments: Arguments,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self,
        *,
        name: str,
        raw: bytes,
        _utf8_clean: str,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        name: str,
        raw: bytes,
        arguments: typing.Optional[Arguments] = None,
        _utf8_clean: typing.Optional[str] = None,
    ) -> None:
        self.name = name
        self.raw = raw
        self.utf8_clean = None
        self.decoded = None
        if not _utf8_clean is None:
            # building fake attribute; no decoding
            self.utf8_clean = _utf8_clean
            return
        assert arguments, "Need arguments for proper decoding"
        try:
            utf8_clean = raw.decode()
            if not CTRL.search(utf8_clean):
                self.utf8_clean = utf8_clean
        except Exception:
            # UTF-8 decode error
            pass
        self._try_decode(arguments)

    def _try_decode_sid(self) -> None:
        try:
            self.decoded = _types.sid.parse_raw(self.raw)
        except Exception:
            return

    def _try_decode_uuid(self) -> None:
        try:
            self.decoded = str(uuid.UUID(bytes=self.raw))
        except Exception:
            return

    def _try_decode_timestamp(self, args: Arguments) -> None:
        if self.utf8_clean:
            try:
                date = _types.timestamp.parse(self.utf8_clean)
            except Exception:
                return
            if args.dateonly:
                self.decoded = str(date.date())
            else:
                self.decoded = str(date)

    def _try_decode_uac(self) -> None:
        if self.utf8_clean:
            try:
                self.decoded = _types.uac.parse(self.utf8_clean.strip())
            except Exception:
                return

    def _try_decode(self, args: Arguments) -> None:
        if self.name in ("objectSid",):
            self._try_decode_sid()
        elif self.name in ("msExchMailboxGuid", "objectGUID"):
            self._try_decode_uuid()
        elif self.name in (
            "pwdLastSet",
            "lastLogon",  # DC local attribute, not synced
            "lastLogonTimestamp",  # set and synced across DCs if "more fresh" than msDS-LogonTimeSyncInterval
            "badPasswordTime",
            "accountExpires",
        ):
            self._try_decode_timestamp(args)
        elif self.name == "userAccountControl":
            self._try_decode_uac()

    @property
    def _base64_value(self) -> str:
        return base64.b64encode(self.raw).decode("ascii")

    def print(self) -> None:
        if not self.decoded is None:
            comment = self.utf8_clean
            if comment is None:
                comment = self._base64_value
            print(f"{self.name}: {self.decoded}  # {comment}")
        elif not self.utf8_clean is None:
            print(f"{self.name}: {self.utf8_clean}")
        else:
            print(f"{self.name}:: {self._base64_value}")

    def to_json(self) -> dict[str, typing.Any]:
        item: dict[str, typing.Any] = {}
        b64_value = self._base64_value
        item["binary"] = b64_value
        if not self.utf8_clean is None:
            item["ldif_value"] = self.utf8_clean
        if not self.decoded is None:
            item["human"] = self.decoded
        elif not self.utf8_clean is None:
            item["human"] = self.utf8_clean
        else:
            item["human"] = self._base64_value
            item["human_is_base64"] = True
        return item

    def human(self) -> str:
        if not self.decoded is None:
            return self.decoded
        elif not self.utf8_clean is None:
            return self.utf8_clean
        else:
            return self._base64_value

    @staticmethod
    def fake_attribute(name: str, value: str) -> Attribute:
        return Attribute(
            name=name,
            raw=value.encode(),
            _utf8_clean=value,
        )


@dataclasses.dataclass(slots=True, kw_only=True)
class Decoder:
    arguments: Arguments

    def read(self, *, dn: str, entry: TEntry) -> dict[str, list[Attribute]]:
        # lowercase attribute name in decoded dict. attribute itself still knows original for LDIF output.
        decoded_entry = {
            name.lower(): [Attribute(name=name, raw=raw, arguments=self.arguments) for raw in raw_values]
            for name, raw_values in entry.items()
        }
        if self.arguments.dndomain or self.arguments.dnpath or self.arguments.dnfullpath:
            dninfo = DNInfo(dn=dn)
            if self.arguments.dndomain:
                decoded_entry["dndomain"] = [
                    Attribute.fake_attribute("dndomain", dninfo.domain),
                ]
            if self.arguments.dnpath:
                decoded_entry["dnpath"] = [
                    Attribute.fake_attribute("dnpath", dninfo.path),
                ]
            if self.arguments.dnfullpath:
                decoded_entry["dnfullpath"] = [
                    Attribute.fake_attribute("dnfullpath", dninfo.full_path),
                ]
        return decoded_entry

    def human(self, *, dn: str, entry: TDecoded) -> dict[str, str]:
        emit: dict[str, typing.Any] = dict(dn=dn)
        for name, attrs in entry.items():
            emit[name] = self.arguments.human_separator.join(attr.human() for attr in attrs)
        return emit

    def json(self, *, dn: str, entry: TDecoded) -> dict[str, str]:
        emit: dict[str, typing.Any] = dict(dn=dn)
        for name, attrs in entry.items():
            emit[name] = [attr.to_json() for attr in attrs]
        return emit

    def _emit_json(self, *, dn: str, entry: TDecoded) -> None:
        if self.arguments.human:
            emit = self.human(dn=dn, entry=entry)
        else:
            emit = self.json(dn=dn, entry=entry)
        json.dump(emit, sys.stdout, ensure_ascii=False)
        print()  # terminate output dicts by newline

    def _emit_ldif(self, *, dn: str, entry: TDecoded) -> None:
        print(f"dn: {dn}")
        for attrs in entry.values():
            for attr in attrs:
                attr.print()
        print()  # separate entries with newlines

    def emit(self, *, dn: str, entry: TDecoded) -> None:
        if self.arguments.human or self.arguments.json:
            self._emit_json(dn=dn, entry=entry)
        else:
            self._emit_ldif(dn=dn, entry=entry)

    def handle(self, *, dn: str, entry: TEntry) -> None:
        entry_attrs = self.read(dn=dn, entry=entry)
        self.emit(dn=dn, entry=entry_attrs)
